import dataclasses
import pathlib


@dataclasses.dataclass
class SumoConfigObject(object):
    scenario_name: str
    path_config_dir: pathlib.Path
    config_name: str = 'sumo.cfg'

    def __post_init__(self):
        assert self.path_config_dir.exists(), f'No directory found at {self.path_config_dir}'
        assert self.path_config_dir.joinpath(self.config_name).exists()




import lxml.etree
import re
from typing import Dict, Any
from pathlib import Path
from sumo_docker_pipeline.logger_unit import logger
from sumo_docker_pipeline.commons.sumo_config_obj import SumoConfigObject
from sumo_docker_pipeline.commons.result_module import SumoResultObjects, ResultFile


class BaseController(object):
    def __init__(self,
                 sumo_command: str = "sumo",
                 is_rewrite_windows_path: bool = True):
        self.sumo_command = sumo_command
        self.is_rewrite_windows_path = is_rewrite_windows_path

    def check_connection(self):
        raise NotImplementedError()

    def get_sumo_version(self) -> str:
        raise NotImplementedError()

    @staticmethod
    def extract_output_options(config_file_name: Path) -> Dict[str, ResultFile]:
        """List up files in output directory.

        :param config_file_name:
        :return:
        """
        tree = lxml.etree.parse(config_file_name.open())
        root = tree.getroot()
        r_files_type = []
        for t in root.find('output'):
            element_name = t.tag
            value_name: Dict[str, Any] = t.attrib
            if isinstance(value_name['value'], bool):
                continue
            if value_name['value'] in ('true', 'false'):
                continue
            # end if
            try:
                float(value_name['value'])
                continue
            except ValueError:
                if 'prefix' in element_name:
                    output_dir_name = config_file_name.parent.joinpath(''.join(value_name['value'].split('/')[:-1]))
                else:
                    output_dir_name = config_file_name.parent.joinpath((value_name['value']))
                # end if
                for p_obj in output_dir_name.glob('*'):
                    if p_obj.is_dir():
                        continue
                    # enf if
                    r_files_type.append(ResultFile(p_obj))
                # end for
            # end try
        # end for
        return {f.name_file: f for f in r_files_type}

    @staticmethod
    def check_output_dir(path_working_dir: Path,
                         target_scenario_name: str,
                         result_file_types: Dict[str, ResultFile]):
        """Creates directory for SUMO output. SUMO raises up Exception if there were NOT a directory for outputs.

        Args:
            path_working_dir: Path to a working directory.
            target_scenario_name: any scenario-name
            result_file_types: {name-of-output-file: Path to directory for saving}

        Returns: None
        """
        for tag_name, file_type_obj in result_file_types.items():
            dir_name = file_type_obj.path_file
            # end if
            confirm_target = Path(path_working_dir).joinpath(target_scenario_name).joinpath(dir_name)
            if confirm_target.exists() is False:
                logger.info(f'created a directory {confirm_target}')
                confirm_target.mkdir()
            # end if
        # end for

    @staticmethod
    def extract_output_dir(path_config_file: Path) -> Path:
        with open(path_config_file, 'r') as f:
            tree = lxml.etree.parse(f)
        # end with
        root = tree.getroot()
        output_prefix_element = root.find('output').find('output-prefix')
        assert output_prefix_element is not None, \
            'output-prefix element does not exist in config file. Check your config file.'
        return Path(path_config_file).parent.joinpath(output_prefix_element.attrib['value'])

    @staticmethod
    def rewrite_windows_path(mount_dir_host: str) -> str:
        """rewrite from Windows style into Unix style.
        Ex. C:\\Users\\kensu\\AppData\\Local\\Temp\\tmp3t95r_eb -> /c/Users/kensu/AppData/Local/Temp/tmp3t95r_eb
        """
        unix_form = Path(mount_dir_host).as_posix()
        drive_prefix = re.search(r'^[C-D]\:', str(unix_form))
        if drive_prefix:
            without_prefix = unix_form[drive_prefix.span()[1]:]
            replaced_prefix = '/' + unix_form[drive_prefix.span()[0]:drive_prefix.span()[1]].lower().replace(':', '')
            path_updated = replaced_prefix + without_prefix
        else:
            path_updated = str(unix_form)
        # end if
        return path_updated

    def start_job(self, path_sumo_config: SumoConfigObject) -> SumoResultObjects:
        raise NotImplementedError()



import dataclasses

from pathlib import Path
from typing import Dict, Optional, Type

from sumo_output_parsers.models.matrix import MatrixObject
from sumo_output_parsers.models.parser import ParserClass

from sumo_docker_pipeline.commons.sumo_config_obj import SumoConfigObject
from sumo_docker_pipeline.logger_unit import logger


@dataclasses.dataclass
class ResultFile(object):
    """A handler class that parse output-files of SUMO output.

    Args:
        path_file: pathlib.Path object that leads into output file's path.
        name_file: (optional) file name of the output file.
    """
    path_file: Path
    name_file: Optional[str] = None

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f'ResultFile class for {self.path_file}'

    def __post_init__(self):
        self.name_file = self.path_file.name


@dataclasses.dataclass
class SumoResultObjects(object):
    """A handler class for SUMO's output file(s).

    Args:
        path_output_dir: A directory where your output xmls exist.
        log_message: A log message from SUMO.
        result_files: {output-file-name: `ResultFile`}
    """
    id_scenario: str
    sumo_config_obj: SumoConfigObject
    path_output_dir: Path
    log_message: Optional[str] = None
    result_files: Optional[Dict[str, ResultFile]] = None

    def parse_output(self,
                     output_file_name: str,
                     parser_class: Type[ParserClass],
                     target_element: str):
        """

        Args:
            parser_class:
            target_element:

        Returns:

        """
        assert output_file_name in self.result_files, f'{output_file_name} does not exits in outputs.'
        try:
            parser = parser_class(self.result_files[output_file_name].path_file)
            matrix_obj = parser.xml2matrix(target_element)
            return matrix_obj
        except Exception as e:
            raise Exception(f'unexpected error. Is your parser correct to the file? your-file: {output_file_name}, '
                            f'your parser: {parser_class}')



from pathlib import Path, WindowsPath
import subprocess
import shutil
from sumo_docker_pipeline import static
from sumo_docker_pipeline.logger_unit import logger
from sumo_docker_pipeline.operation_module.base_operation import BaseController
from sumo_docker_pipeline.commons.sumo_config_obj import SumoConfigObject
from sumo_docker_pipeline.commons.result_module import SumoResultObjects


class LocalSumoController(BaseController):
    def __init__(self,
                 sumo_command: str = None,
                 is_rewrite_windows_path: bool = True):
        if sumo_command is None:
            sumo_command = shutil.which('sumo')
            if sumo_command is None:
                raise Exception('command `sumo` is NOT found in your system. Check your environment. '
                                'or, give path to sumo manually. Use the `sumo_command` argument.')

        super(LocalSumoController, self).__init__(
            sumo_command=sumo_command,
            is_rewrite_windows_path=is_rewrite_windows_path)
        self.check_connection()



    def check_connection(self):
        sumo_bash_command = [self.sumo_command]
        pipe_obj = subprocess.Popen(sumo_bash_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        outs, errs = pipe_obj.communicate()
        r_code = pipe_obj.returncode
        assert r_code == 0
        assert "German Aerospace Center" in outs.decode('utf-8')

    def get_sumo_version(self) -> str:
        sumo_bash_command = [self.sumo_command, '-V']
        pipe_obj = subprocess.Popen(sumo_bash_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        outs, errs = pipe_obj.communicate()
        r_code = pipe_obj.returncode
        assert r_code == 0
        return outs.decode('utf-8')

    def start_job(self, sumo_config: SumoConfigObject) -> SumoResultObjects:
        """Run SUMO on local.

        Args:
            target_scenario_name: Nothing. Keep it None.
            config_file_name: "sumo.cfg" name.

        Returns: `SumoResultObjects`
        """
        if self.is_rewrite_windows_path and isinstance(sumo_config.path_config_dir, WindowsPath):
            # If windows...Path structure is broken. Fix it manually.
            path_config_file = sumo_config.path_config_dir.as_posix()
        else:
            path_config_file = sumo_config.path_config_dir.joinpath(sumo_config.config_name)
        # end if
        job_command = f'{self.sumo_command} -c {path_config_file}'
        logger.debug(f'executing job with command {job_command}')

        sumo_bash_command = [self.sumo_command, '-c', path_config_file]
        pipe_obj = subprocess.Popen(sumo_bash_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        outs, errs = pipe_obj.communicate()
        r_code = pipe_obj.returncode
        assert r_code == 0

        res_obj = SumoResultObjects(
            id_scenario=sumo_config.scenario_name,
            sumo_config_obj=sumo_config,
            log_message=outs.decode('utf-8'),
            path_output_dir=self.extract_output_dir(sumo_config.path_config_dir.joinpath(sumo_config.config_name)))
        return res_obj



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sumo_docker_pipeline',
 'sumo_docker_pipeline.commons',
 'sumo_docker_pipeline.operation_module']

package_data = \
{'': ['*']}

install_requires = \
['bs4', 'docker', 'joblib', 'lxml', 'pandas', 'sumo-output-parsers>=0.5,<0.6']

extras_require = \
{'full': ['Shapely>=1.7.0,<2.0.0',
          'pyproj>=3.0.0,<4.0.0',
          'SumoNetVis>=1.6.0,<2.0.0',
          'geopandas>=0.10.0,<0.11.0',
          'geoviews>=1.9.1,<2.0.0']}

setup_kwargs = {
    'name': 'sumo-docker-pipeline',
    'version': '2.0',
    'description': 'A pipeline to call a traffic simulator: SUMO',
    'long_description': '# sumo_docker_pipeline\n- - -\n\nThe package `sumo_docker_pipeline` enables you to run a traffic simulator [SUMO](https://sumo.dlr.de/docs/index.html) efficiently \nand to interact with Python easily. \nThe package is valid when you need to run SUMO iteratively.\n\nSUMO is often tricky to install locally because of its dependencies. \nThus, it\'s a straightforward idea to run SUMO in a docker container.\n\nHowever, another issue arises when we run SUMO in a docker container. \nIt is challenging to construct a pipeline between SUMO and API.\n\nThe package provides an easy-to-use API; \nat the same time, \nSUMO runs in a docker container.\n\n# Requirement\n\n- python > 3.5\n- docker \n- docker-compose\n\n# Install\n\n## Pull the image (or build of a docker image with SUMO)\n\nThe existing image is on the [Dockerhub](https://hub.docker.com/repository/docker/kensukemi/sumo-ubuntu18).\n\n```shell\ndocker pull kensukemi/sumo-ubuntu18\n```\n\nIf you prefer to build with yourself, you run the following command.\n\n```shell\ndocker-compose build \n```\n\n## Install a python package\n\n```shell\nmake install\n```\n\n# Example case: iterative run with parameter updates\n\nLet\'s say that you want to run SUMO iteratively.\nAt the same time, you want to change input parameters depending on the results of a simulation. \n\nIn that case, you need to check the output of SUMO and update the parameters.\n`sumo_docker_pipeline` package enables you to make the process automatic.\n\n[![](https://user-images.githubusercontent.com/1772712/119264146-34563500-bbe2-11eb-9288-2e4e841ff803.png)]()\n\n## Setups\n\n1. creation of a directory where you save SUMO\'s configuration.\n2. creation of template-files of SUMO\'s configuration.\n3. running the pipeline.\n\n## 1. creation of a directory\n\nIt is a directory that SUMO accesses.\nLet\'s say that we create `test/resources/config_template`\n\n## 2. creation of SUMO\'s configuration\n\nYou prepare configuration files which SUMO requires.\nThe format of the conf. files are totally same as SUMO\'s requirements.\n\nThe only difference is that you write wildcard `?` at the place where you wanna replace during pipeline.\n\nFor example, `tests/resources/config_template/grid.flows.xml` has the following element,\n\n```xml\n<vType vClass="passenger" id="passenger"  tau="0.5" speedDev="0.1" maxSpeed="?" minGap="?" accel="?" decel="?" latAlignment="center" />\n```\n\nWith the `sumo_docker_pipeline` package, you can replace the attributes with the wildcards `?`.\n\n## 3. running the pipeline\n\nSee `examples` directory.\n\n\n\n# For developer\n\n```shell\npytest tests\n```\n\n# license and credit\n\nThe source code is licensed MIT. The website content is licensed CC BY 4.0.\n\n\n```\n@misc{sumo-docker-pipeline,\n  author = {Kensuke Mitsuzawa},\n  title = {sumo-docker-pipeline},\n  year = {2021},\n  publisher = {GitHub},\n  journal = {GitHub repository},\n  howpublished = {\\url{https://github.com/Kensuke-Mitsuzawa/sumo_docker_pipeline}},\n}\n```',
    'author': 'Kensuke Mitsuzawa',
    'author_email': 'kensuke.mit@gmail.com',
    'maintainer': 'Kensuke Mitsuzawa',
    'maintainer_email': 'kensuke.mit@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from sumo_docker_pipeline.pipeline import DockerPipeline, LocalSumoPipeline
from sumo_docker_pipeline.operation_module import SumoDockerController, LocalSumoController
from sumo_docker_pipeline.commons.result_module import ResultFile, SumoResultObjects
from .commons import SumoConfigObject
from .config_generation_module import Template2SuMoConfig



import typing
import joblib
from datetime import datetime
from pathlib import Path
from typing import Optional
from sumo_docker_pipeline.logger_unit import logger
from sumo_docker_pipeline.operation_module.docker_operation_module import SumoDockerController
from sumo_docker_pipeline.operation_module.local_operation_module import LocalSumoController
from sumo_docker_pipeline.commons.result_module import SumoResultObjects
from sumo_docker_pipeline.commons.sumo_config_obj import SumoConfigObject
import time


class BasePipeline(object):
    def __init__(self,
                 path_working_dir: typing.Optional[Path] = None,
                 n_jobs: int = 1):
        self.n_jobs = n_jobs
        if path_working_dir is None:
            self.path_working_dir = Path('/tmp').joinpath('sumo_docker_pipeline').absolute()
        else:
            self.path_working_dir = path_working_dir.absolute()
        # end if

    def get_data_directory(self) -> Path:
        raise NotImplementedError()

    def run_simulation(self,
                       sumo_configs: typing.List[SumoConfigObject],
                       is_overwrite: bool = False) -> SumoResultObjects:
        raise NotImplementedError()


class LocalSumoPipeline(BasePipeline):
    def __init__(self,
                 is_rewrite_windows_path: bool = True,
                 path_working_dir: Path = None,
                 n_jobs: int = 1,
                 sumo_command: str = '/bin/sumo'):
        """A pipeline interface to run SUMO-docker.

        Args:
            path_working_dir: a path to save tmp files.
            n_jobs: the number of cores.
        """
        super(LocalSumoPipeline, self).__init__(path_working_dir=path_working_dir, n_jobs=n_jobs)
        self.sumo_command = sumo_command

        # todo move Template2SuMoConfig to other module.
        # self.template_generator = Template2SuMoConfig(path_config_file=str(path_config_file),
        #                                               path_destination_dir=str(path_destination_scenario))
        self.is_rewrite_windows_path = is_rewrite_windows_path

    def one_simulation(self, sumo_config_object: SumoConfigObject) -> typing.Tuple[str, SumoResultObjects]:
        sumo_controller = LocalSumoController(sumo_command=self.sumo_command)
        sumo_result_obj = sumo_controller.start_job(sumo_config=sumo_config_object)
        return sumo_config_object.scenario_name, sumo_result_obj

    def run_simulation(self,
                       sumo_configs: typing.List[SumoConfigObject],
                       is_overwrite: bool = False) -> typing.Dict[str, SumoResultObjects]:
        """Run SUMO simulation.

        Args:
            sumo_configs: List of SUMO Config objects.
            is_overwrite: True, then the method overwrites outputs from SUMO. False raises Exception if there is a destination directory already. Default False.

        Returns: {scenario-name: `SumoResultObjects`}
        """
        logger.info(f'running sumo simulator now...')
        sumo_result_objects = joblib.Parallel(n_jobs=self.n_jobs)(joblib.delayed(
            self.one_simulation)(conf) for conf in sumo_configs)
        logger.info(f'done the simulation.')
        _ = dict(sumo_result_objects)
        return _


class DockerPipeline(BasePipeline):
    def __init__(self,
                 path_mount_working_dir: Optional[Path] = None,
                 docker_image_name: str = 'kensukemi/sumo-ubuntu18',
                 is_rewrite_windows_path: bool = True,
                 n_jobs: int = 1,
                 time_interval_future_check: float = 3.0,
                 limit_max_wait: float = 3600):
        """A pipeline interface to run SUMO-docker.

        Args:
            The other config files should be in the same directory (or under the sub-directory)
            path_mount_working_dir: A path to directory where a container mount as the shared directory.
            docker_image_name: A name of docker-image that you call.
            is_rewrite_windows_path: True, then the class updates Path only when your OS is Windows.
            n_jobs: the number of parallel computations.
            time_interval_future_check: Time interval to check Task status.
            limit_max_wait: Time limit (seconds) to force end process.
        """
        super(DockerPipeline, self).__init__(path_working_dir=path_mount_working_dir, n_jobs=n_jobs)
        self.path_mount_working_dir = self.path_working_dir
        self.docker_image_name = docker_image_name
        self.is_rewrite_windows_path = is_rewrite_windows_path
        self.time_interval_future_check = time_interval_future_check
        self.limit_max_wait = limit_max_wait

    def get_data_directory(self) -> Path:
        return self.path_mount_working_dir

    def one_job(self, sumo_config_obj: SumoConfigObject) -> SumoResultObjects:
        logger.debug(f'running sumo simulator now...')
        time_stamp_current = datetime.now().timestamp()
        sumo_controller = SumoDockerController(
            container_name_base=f'sumo-docker-{sumo_config_obj.scenario_name}-{time_stamp_current}',
            image_name=self.docker_image_name,
        )
        sumo_result_obj = sumo_controller.start_job(sumo_config=sumo_config_obj)
        logger.debug(f'done the simulation.')
        return sumo_result_obj

    def run_simulation(self,
                       sumo_configs: typing.List[SumoConfigObject],
                       is_overwrite: bool = False) -> typing.Dict[str, SumoResultObjects]:
        """Run SUMO simulation in a docker container.

        Args:
            sumo_configs: List of SumoConfigObject.
            is_overwrite: True, then the method overwrites outputs from SUMO.
            False raises Exception if there is a destination directory already. Default False.
        Returns:
            dict {scenario-name: `SumoResultObjects`}
        """
        logger.info(f'making the new config files')
        from concurrent.futures import ThreadPoolExecutor
        pool = ThreadPoolExecutor(self.n_jobs)
        s_future_pool = []
        logger.debug(f'starting tasks...')
        for conf in sumo_configs:
            future = pool.submit(self.one_job, (conf))
            s_future_pool.append(future)
        # end for
        logger.debug(f'submitted all tasks.')

        time_at_start = datetime.now()
        logger.debug(f'waiting for task-ends...')
        while True:
            is_end_all = all(f.done() is True for f in s_future_pool)
            if is_end_all:
                break
            time.sleep(self.time_interval_future_check)
            if (datetime.now() - time_at_start).total_seconds() > self.limit_max_wait:
                raise TimeoutError(f'We waited {self.limit_max_wait} seconds. Not finished yet.')
            # end if
        # end while

        d_scenario_name2result = {}
        for f_obj in s_future_pool:
            r: SumoResultObjects = f_obj.result()
            d_scenario_name2result[r.id_scenario] = r
        # end for
        return d_scenario_name2result
